"""

Metta sonobuoy CLI plugin.

Provides functionality to manually run and inspect sonobuoy jobs

"""
import logging
import time

from mirantis.testing.metta.environment import Environment
from mirantis.testing.metta.workload import METTA_PLUGIN_TYPE_WORKLOAD
from mirantis.testing.metta_cli.base import CliBase, cli_output

from .sonobuoy import METTA_PLUGIN_ID_SONOBUOY_WORKLOAD, Status

logger = logging.getLogger('metta.cli.sonobuoy')

METTA_PLUGIN_ID_SONOBUOY_CLI = 'metta_sonobuoy_cli'
""" cli plugin_id for the sonobuoy plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class SonobuoyCliPlugin(CliBase):
    """Fire command/group generator for sonobuoy commands."""

    def fire(self):
        """Return CLI Command group."""
        if self.environment.fixtures.get(plugin_type=METTA_PLUGIN_TYPE_WORKLOAD,
                                         plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD,
                                         exception_if_missing=False) is not None:
            return {
                'contrib': {
                    'sonobuoy': SonobuoyGroup(self.environment)
                }
            }

        return {}


class SonobuoyGroup():
    """Base Fire command group for sonobuoy cli commands."""

    def __init__(self, environment: Environment):
        """Inject Environment into command group."""
        self.environment = environment

    def _select_fixture(self, instance_id: str = ''):
        """Pick a matching workload plugin."""
        try:
            if instance_id:
                return self.environment.fixtures.get(
                    plugin_type=METTA_PLUGIN_TYPE_WORKLOAD,
                    plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD, instance_id=instance_id)

            # Get the highest priority provisioner
            return self.environment.fixtures.get(
                plugin_type=METTA_PLUGIN_TYPE_WORKLOAD, plugin_id=METTA_PLUGIN_ID_SONOBUOY_WORKLOAD)

        except KeyError as err:
            raise ValueError("No usable kubernetes client was found for sonobuoy"
                             "to pull a kubeconfig from") from err

    def _select_instance(self, instance_id: str = ''):
        """Create a sonobuoy workload plugin instance."""
        fixture = self._select_fixture(instance_id=instance_id)
        plugin = fixture.plugin
        # @TODO allow filtering of kubernetes client instances
        instance = plugin.create_instance(self.environment.fixtures)
        return instance

    def info(self, instance_id: str = '', deep: bool = False):
        """Get info about a provisioner plugin."""
        fixture = self._select_fixture(instance_id=instance_id)

        info = {
            'fixture': {
                'plugin_type': fixture.plugin_type,
                'plugin_id': fixture.plugin_id,
                'instance_id': fixture.instance_id,
                'priority': fixture.priority,
            }
        }

        if deep:
            if hasattr(fixture.plugin, 'info'):
                info.update(fixture.plugin.info(True))

        return cli_output(info)

    def status(self, instance_id: str = ''):
        """Get active sonobuoy status."""
        instance = self._select_instance(instance_id=instance_id)
        status = instance.status()

        if status is None:
            status_info = {
                'status': "None",
                'plugins': {}
            }
        else:
            status_info = {
                'status': status.status.value,
                'plugins': {plugin_id: status.plugin(plugin_id) for
                            plugin_id in status.plugin_list()}
            }

        return cli_output(status_info)

    # pylint: disable=protected-access
    def crb(self, instance_id: str = '', remove: bool = False):
        """Create the crb needed to run sonobuoy."""
        instance = self._select_instance(instance_id=instance_id)

        if remove:
            instance._delete_k8s_crb()
        else:
            instance._create_k8s_crb()

    def run(self, instance_id: str = '', wait: bool = False):
        """Run sonobuoy workload."""
        instance = self._select_instance(instance_id=instance_id)
        instance.apply(wait=wait)

    def wait(self, instance_id: str = '', step: int = 5, limit: int = 1000):
        """Wait until no longer running."""
        instance = self._select_instance(instance_id=instance_id)
        print('{')
        for i in range(0, limit):
            status = instance.status()
            if status is None:
                status_info = {
                    'status': "None",
                    'plugins': {}
                }
            else:
                status_info = {
                    'status': status.status.value,
                    'plugins': {plugin_id: status.plugin(plugin_id) for
                                plugin_id in status.plugin_list()}
                }

            print(f"{i}: {status_info},")
            if status is None or status.status not in [Status.RUNNING]:
                break

            time.sleep(step)
        print('}')

    def destroy(self, instance_id: str = '', wait: bool = False):
        """Remove all sonobuoy infrastructure."""
        instance = self._select_instance(instance_id=instance_id)
        instance.destroy(wait=wait)

    def logs(self, instance_id: str = '', follow: bool = False):
        """Retrieve sonobuoy logs."""
        instance = self._select_instance(instance_id=instance_id)
        instance.logs(follow=follow)

    def retrieve(self, instance_id: str = ''):
        """Retrieve the results from the sonobuoy workload instance."""
        instance = self._select_instance(instance_id=instance_id)
        try:
            instance.retrieve()

        # broad catch to allow message formatting for cli output
        # pylint: disable=broad-except
        except Exception as err:
            logger.error("Retrieve failed: %s", err)



"""

METTA PLUGIN: Text output plugin.

An output plugin which holds a single string value as its content.

"""

import logging


logger = logging.getLogger('metta.contrib.common.output.text')

METTA_PLUGIN_ID_OUTPUT_TEXT = 'text'
""" output plugin_id for the text plugin """


class TextOutputPlugin:
    """Metta Output plugin a text output type.

    this just gets and sets text. Nothing fancy.

    """

    def __init__(self, environment, instance_id, text: str = ''):
        """Run the super constructor but also set class properties.

        Parameters:
        -----------
        data (str) : any string data to be stored

        Raises:
        -------
        A configerus.validate.ValidationError is raised if you passed in a
        validator target and validation failed.

        An AssertionError is raised if you didn't pass in a Dict.

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        self.set_text(text)

    def set_text(self, data: str):
        """Overwrite the contents of the output string."""
        self.text = data

    def get_output(self) -> str:
        """Retrieve assigned output string.

        Returns:
        --------
        The string that was assigned using .arguments() or set_text()

        Raises:
        -------
        AttributeError if you tried to get text before you have assigned it.

        """
        return self.text

    def info(self):
        """Return dict data about this plugin for introspection."""
        return {
            'output': {
                'text': self.text
            }
        }



"""

Metta Client plugin for a docker client using docker-py.

"""
import logging
import os

from docker import DockerClient

logger = logging.getLogger('metta.contrib.docker.client.dockerpy')

METTA_PLUGIN_ID_DOCKER_CLIENT = 'metta_docker'
""" client plugin_id for the metta dummy plugin """


# this interface is common for all Metta plugins, but CLI plugins underuse it
# pylint: disable=too-few-public-methods
class DockerPyClientPlugin(DockerClient):
    """Metta Client plugin for docker using the docker-py library.

    Most of the functionality is provided by the docker-py::DockerClient class,
    while the metta plugin side is respondible for configuring the client.

    """

    # This is really what it takes to configure both the plugin and the client
    # pylint: disable=too-many-arguments
    def __init__(self, environment, instance_id, host: str, cert_path: str, tls_verify: bool = True,
                 compose_tls_version: str = 'TLSv1_2', version: str = 'auto'):
        """Set class properties.

        In order to decorate this existing class as a DockerClient, without
        using the DockerClient constructor, nor the from_env class method, we
        have to reproduce the APIClient construction.
        We could rewrite some of the DockerClient functionality, or we can just
        create a throwaway DockerClient instance and steal its ApiClient and
        use it.

        This also lets us easily include any other env variables that might be
        in scope.

        @Note that we don't run the docker-py constructor as we build our own
            client.

        Parameters:
        -----------
        All of the arguments are handed off to the DockerClient.from_env class
        method. All arguments are converted to the related ENV equivalent if it
        had been set outside of this python code.

        host (str) [DOCKER_HOST] what daemon host to use (docker socket)

        cert_path (str) [DOCKER_CERT_PATH] path to rsa keys for authorization

        tls_verify (bool) [DOCKER_TLS_VERIFY] should the client pursue TLS
            verification

        compose_tls_version (str) [COMPOSE_TLS_VERSION] what TLS version should
            the Docker client use for docker compose.

        """
        self.environment = environment
        """ Environemnt in which this plugin exists """
        self.instance_id = instance_id
        """ Unique id for this plugin instance """

        logger.debug("Configuring docker client with args for host: %s", host)

        self.host = host
        self.cert_path = cert_path
        self.tls_verify = '1' if tls_verify else '0'
        self.compose_tls_version = compose_tls_version

        env = os.environ.copy()
        env['DOCKER_HOST'] = self.host
        env['DOCKER_CERT_PATH'] = self.cert_path
        env['DOCKER_TLS_VERIFY'] = self.tls_verify
        env['COMPOSE_TLS_VERSION'] = self.compose_tls_version

        # Build a client in the classical way, but we just take it's api and
        # throw the client away
        throwaway = DockerClient.from_env(environment=env, version=version)
        self.api = throwaway.api

    def info(self):
        """Return dict data about this plugin for introspection."""
        return {
            'docker': {
                'host': self.host,
                'cert_path': self.cert_path,
                'tls_verify': self.tls_verify,
                'compose_tls_version': self.compose_tls_version
            }
        }
